// Copyright 2018-2023 the Deno authors. All rights reserved. MIT license.

use deno_core::error::AnyError;
use tower_lsp::LspService;
use tower_lsp::Server;

use crate::lsp::language_server::LanguageServer;
pub use repl::ReplCompletionItem;
pub use repl::ReplLanguageServer;

use self::diagnostics::should_send_diagnostic_batch_index_notifications;

mod analysis;
mod cache;
mod capabilities;
mod client;
mod code_lens;
mod completions;
mod config;
mod diagnostics;
mod documents;
pub mod language_server;
mod logging;
mod lsp_custom;
mod npm;
mod parent_process_checker;
mod path_to_regex;
mod performance;
mod refactor;
mod registries;
mod repl;
mod semantic_tokens;
mod testing;
mod text;
mod tsc;
mod urls;

pub async fn start() -> Result<(), AnyError> {
  let stdin = tokio::io::stdin();
  let stdout = tokio::io::stdout();

  let builder = LspService::build(|client| {
    language_server::LanguageServer::new(client::Client::from_tower(client))
  })
  // TODO(nayeemrmn): The extension has replaced this with the `deno.cache`
  // command as of vscode_deno 3.21.0 / 2023.09.05. Remove this eventually.
  .custom_method(lsp_custom::CACHE_REQUEST, LanguageServer::cache_request)
  .custom_method(
    lsp_custom::PERFORMANCE_REQUEST,
    LanguageServer::performance_request,
  )
  // TODO(nayeemrmn): The extension has replaced this with the
  // `deno.reloadImportRegistries` command as of vscode_deno
  // 3.26.0 / 2023.10.10. Remove this eventually.
  .custom_method(
    lsp_custom::RELOAD_IMPORT_REGISTRIES_REQUEST,
    LanguageServer::reload_import_registries_request,
  )
  .custom_method(lsp_custom::TASK_REQUEST, LanguageServer::task_definitions)
  // TODO(nayeemrmn): Rename this to `deno/taskDefinitions` in vscode_deno and
  // remove this alias.
  .custom_method("deno/task", LanguageServer::task_definitions)
  .custom_method(testing::TEST_RUN_REQUEST, LanguageServer::test_run_request)
  .custom_method(
    testing::TEST_RUN_CANCEL_REQUEST,
    LanguageServer::test_run_cancel_request,
  )
  .custom_method(
    lsp_custom::VIRTUAL_TEXT_DOCUMENT,
    LanguageServer::virtual_text_document,
  );

  let builder = if should_send_diagnostic_batch_index_notifications() {
    builder.custom_method(
      lsp_custom::LATEST_DIAGNOSTIC_BATCH_INDEX,
      LanguageServer::latest_diagnostic_batch_index_request,
    )
  } else {
    builder
  };

  let (service, socket) = builder.finish();

  Server::new(stdin, stdout, socket).serve(service).await;

  Ok(())
}
